#pragma once
#include <cstdint>
#include <vector>
#include "Event.h"
#include <map>
#include <thread>
#include "Key.hpp"
#include "rewindow/Cursors.hpp"
#include "MouseButton.hpp"
#include "GamepadButton.hpp"
#include "J3ML/LinearAlgebra.hpp"
#include "WindowEvents.hpp"
#include <format>
#include <queue>

#include <experimental/propagate_const>


namespace ReWindow {
    enum class RenderingAPI: uint8_t {
        OPENGL = 0,
        //Vulkan is unimplemented.
        VULKAN = 1,
    };
    //

    class RWindowImpl;
    class XLibImpl;
    class Win32Impl;
    class RWindow;
}

class ReWindow::RWindow {
private:
    RWindowImpl* Impl = nullptr;
public:
    //RWindow() = default;
    //RWindow();

    static RWindowImpl* GetPlatformImplementation(const RWindowParams& args) {

#if UNIX
        return new XLibImpl(args);
#endif

#if WIN32
        return new Win32Impl(args);
#endif
        return nullptr;
    }


    explicit RWindow(const RWindowParams& params)
        : Impl(GetPlatformImplementation(params)) { }

    RWindow()
    : RWindow({
        .title = "Redacted Window",
        .width = 640, .height = 480,
        .fullscreen = false,
        .resizable = true,
        .vsync = false
    }) { }

    // TODO: Is this one valid?
    explicit RWindow(RWindowImpl* wImpl)
        : Impl(wImpl) { }

    RWindow(RWindowImpl* wImpl, RenderImpl* wRenderImpl)
        : Impl(wImpl), Renderer(wRenderImpl) { }

    /*
    explicit RWindow(const std::string& wTitle,
                     int wWidth = 640, int wHeight = 480,
                     RenderingAPI wRenderer = RenderingAPI::OPENGL,
                     bool wFullscreen = false,
                     bool wResizable = true,
                     bool wVsync = false) :
                     RWindowImpl(wTitle, wWidth, wHeight, wRenderer, wFullscreen, wResizable, wVsync) {};
                     */
    ~RWindow() = default;
public:
    // Platform dependant
    void Open() { Impl->Open(); };
    void Close() { Impl->Close(); };
    void PollEvents() { Impl->PollEvents(); };
    void Refresh() { Impl->Refresh(); };
public:
    // Shared
    /// Returns whether the window currently has mouse and/or keyboard focus.
    [[nodiscard]] bool IsFocused() const { return Impl->IsFocused(); };
    /// Returns whether the window is currently in Fullscreen.
    // TODO: Support Fullscreen, FullscreenWindowed, and Windowed?
    [[nodiscard]] bool IsFullscreen() const { return Impl->IsFullscreen(); } ;
    /// Returns whether the window can be resized.
    [[nodiscard]] bool IsResizable() const { return Impl->IsResizable(); } ;
    /// Returns whether V-Sync is enabled.
    [[nodiscard]] bool IsVsyncEnabled() const { return Impl->IsVsyncEnabled(); } ;
    /// Returns whether the window is considered to be alive. Once dead, any logic loop should be terminated, and the cleanup procedure should run.
    [[nodiscard]] bool IsAlive() const { return Impl->IsAlive(); } ;
    // Should have IsOpen since window could be closed then reopened
public:
    // Platform dependant
    /// Sets whether or not to make the window fullscreen.
    /// @note This is implemented per-OS, and as such, it simply requests the OS to do what we want. No guarantee about follow-through can be given.
    void SetFullscreen(bool fs) { Impl->SetFullscreen(fs); };
    /// Sets whether or not to make the window resizable.
    /// @note This is implemented per-OS, and as such, it simply requests the OS to do what we want. No guarantee about follow-through can be given.
    void SetResizable(bool resizable) { Impl->SetResizable(resizable); };
    /// Sets whether or not to enable vertical synchronization.
    /// @note This is implemented per-OS, and as such, it simply requests the OS to do what we want. No guarantee about follow-through can be given.
    void SetVsyncEnabled(bool vsync) { Impl->SetVsyncEnabled(vsync); };
    /// Requests the operating system to change the window size.
    /// @param width
    /// @param height
    void SetSize(int width, int height) { Impl->SetSize(width, height); };
    /// Requests the operating system to change the window size.
    /// @param size
    void SetSize(const Vector2& size) { Impl->SetSize(size); };

    // Shared
    /// Sets the title of this window.
    void SetTitle(const std::string& title) { Impl->SetTitle(title); };
public:
    // Shared
    [[nodiscard]] std::string GetTitle() const { return Impl->GetTitle(); } ;
    /// Returns the position of the window's top-left corner relative to the display
    [[nodiscard]] Vector2 GetPos() const { return Impl->GetPos(); } ;
    /// Returns the known size of the window, in {x,y} pixel measurement.
    [[nodiscard]] Vector2 GetSize() const { return Impl->GetSize(); } ;
    /// Returns the horizontal length of the renderable area in pixels.
    [[nodiscard]] int GetWidth() const { return Impl->GetWidth(); } ;
    /// Returns the vertical length of the renderable area, in pixels.
    [[nodiscard]] int GetHeight() const { return Impl->GetHeight(); } ;
    /// Returns the amount of time, in seconds, between the current and last frame.
    /// Technically, no, it returns the elapsed time of the frame, start to finish.
    [[nodiscard]] float GetDeltaTime() const { return Impl->GetDeltaTime(); } ;
    /// Returns the approximate frames-per-second using delta time.
    [[nodiscard]] float GetRefreshRate() const { return Impl->GetRefreshRate(); } ;
    /// Returns the number of frames ran since the windows' creation.
    [[nodiscard]] float GetRefreshCounter() const { return Impl->GetRefreshCounter(); } ;
public:
    // Figure out what to do with these later
    void Raise() const { Impl->Raise(); };
    void Lower() const { Impl->Lower(); };
    void DestroyOSWindowHandle() { Impl->DestroyOSWindowHandle(); };
    void SetCursorVisible(bool cursor_enable) { Impl->SetCursorVisible(cursor_enable); };
    Vector2 GetAccurateMouseCoordinates() const { return Impl->GetAccurateMouseCoordinates(); } ;
    void GLSwapBuffers() { Impl->GLSwapBuffers(); };
    void Fullscreen() { Impl->Fullscreen(); };
    void RestoreFromFullscreen() { Impl->RestoreFromFullscreen(); };
    // I know this doesn't modify the class, but it indirectly modifies the window
    // Making it const just seems deceptive.
    void SetCursorStyle(CursorStyle style) const { return Impl->SetCursorStyle(style); } ;
    Vector2 GetPositionOfRenderableArea() const { return Impl->GetPositionOfRenderableArea(); } ;
    std::string getGraphicsDriverVendor() { return Impl->getGraphicsDriverVendor(); };
    void SetFlag(RWindowFlags flag, bool state) { Impl->SetFlag(flag, state); };
    void processKeyRelease (Key key) { Impl->processKeyRelease(key); };
    void processKeyPress (Key key) { Impl->processKeyPress(key); };
    void processOnClose() { Impl->processOnClose(); };
    void processOnOpen() { Impl->processOnOpen(); };
    void processMousePress(const MouseButton& btn) { Impl->processMousePress(btn); };
    void processMouseRelease(const MouseButton& btn) { Impl->processMouseRelease(btn); };
    void processFocusIn() { Impl->processFocusIn(); };
    void processFocusOut() { Impl->processFocusOut(); };
    void processMouseMove(Vector2 last_pos, Vector2 new_pos) { Impl->processMouseMove(last_pos, new_pos); };
    void processMouseWheel(int scrolls) { Impl->processMouseWheel(scrolls); };
    virtual void OnOpen() {}
    virtual void OnClosing() {}
    virtual void OnFocusLost(const RWindowEvent& e) {}
    virtual void OnFocusGain(const RWindowEvent& e) {}
    virtual void OnRefresh(float elapsed) {}
    virtual void OnResizeSuccess() {}
    virtual bool OnResizeRequest(const WindowResizeRequestEvent& e) { return true;}
    virtual void OnKeyDown(const KeyDownEvent&) {}
    virtual void OnKeyUp(const KeyUpEvent&) {}
    virtual void OnMouseMove(const MouseMoveEvent&) {}
    virtual void OnMouseButtonDown(const MouseButtonDownEvent&) {}
    virtual void OnMouseButtonUp(const MouseButtonUpEvent&) {}
    virtual void OnMouseWheel(const MouseWheelEvent&) {}
    Event<> OnOpenEvent;
    Event<> OnClosingEvent;
    Event<RWindowEvent> OnFocusLostEvent;
    Event<RWindowEvent> OnFocusGainEvent;
    Event<float> OnRefreshEvent;
    Event<WindowResizeRequestEvent> OnResizeRequestEvent;
    Event<KeyDownEvent> OnKeyDownEvent;
    Event<KeyUpEvent> OnKeyUpEvent;
    Event<MouseMoveEvent> OnMouseMoveEvent;
    Event<MouseButtonDownEvent> OnMouseButtonDownEvent;
    Event<MouseButtonUpEvent> OnMouseButtonUpEvent;
    Event<MouseWheelEvent> OnMouseWheelEvent;

    Vector2 GetMouseCoordinates() const { return Impl->GetMouseCoordinates(); };
    bool GetFlag(RWindowFlags flag) const { return Impl->GetFlag(flag); };
    //bool IsAlive() const { return Impl->
    //IsAlive(); };
    void SetSizeWithoutEvent(const Vector2& size) { Impl->SetSizeWithoutEvent(size); };
    void LogEvent(const RWindowEvent& e) { Impl->LogEvent(e); };
    RWindowEvent GetLastEvent() const { return Impl->GetLastEvent(); }
    void SetLastKnownWindowSize(const Vector2& size) { Impl->SetLastKnownWindowSize(size); };
    void SetRenderer(RenderingAPI api) { Impl->SetRenderer(api); };
    bool IsKeyDown(Key key) const { return Impl->IsKeyDown(key); };
    bool IsMouseButtonDown(const MouseButton &button) const { return Impl->IsMouseButtonDown(button); };
    void ManagedRefresh() { Impl->ManagedRefresh(); };
    float ComputeElapsedFrameTimeSeconds(std::chrono::steady_clock::time_point start, std::chrono::steady_clock::time_point end) { return Impl->ComputeElapsedFrameTimeSeconds(start, end); };
    std::chrono::steady_clock::time_point GetTimestamp() { return Impl->GetTimestamp(); };
    void UpdateFrameTiming(float frame_time) { Impl->UpdateFrameTiming(frame_time); };
    //bool IsResizable() const { return Impl->
    //IsResizable(); };
    //bool IsFullscreen() const { return Impl->
    //IsFullscreen(); };
    //bool IsFocused() const { return Impl->
    //IsFocused(); };
    //bool IsVsyncEnabled() const { return Impl->
    //IsVsyncEnabled(); };
    void ForceClose() { Impl->ForceClose(); };
    void ForceCloseAndTerminateProgram() { Impl->ForceCloseAndTerminateProgram(); };
    int GetMouseWheelPersistent() const { return Impl->GetMouseWheelPersistent();}
    [[nodiscard]] bool IsOpen() const { return Impl->IsOpen();}
    [[nodiscard]] bool IsClosing() const { return Impl->IsClosing();}

};